@@ -1,4 +1,5 @@
 /* identify.c - by Mark Lord (C) 2000-2007 -- freely distributable */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -10,7 +11,6 @@
 #define __USE_XOPEN
 #endif
 
-#include <unistd.h>
 #include "hdparm.h"
 
 /* device types */
@@ -37,7 +37,7 @@
 #define LENGTH_FW_REV		 4  /*  4 words (8 bytes or characters) */
 #define START_MODEL    		27  /* ASCII model number */
 #define LENGTH_MODEL    	20  /* 20 words (40 bytes or characters) */
-#define SECTOR_XFER_MAX	        47  /* r/w multiple: max sectors xfered */
+#define SECTOR_XFER_MAX		47  /* r/w multiple: max sectors xfered */
 #define DWORD_IO		48  /* can do double-word IO (ATA-1 only) */
 #define CAPAB_0			49  /* capabilities */
 #define CAPAB_1			50
@@ -46,7 +46,7 @@
 #define WHATS_VALID		53  /* what fields are valid */
 #define LCYLS_CUR		54  /* current logical cylinders */
 #define LHEADS_CUR		55  /* current logical heads */
-#define LSECTS_CUR	        56  /* current logical sectors/track */
+#define LSECTS_CUR		56  /* current logical sectors/track */
 #define CAPACITY_LSB		57  /* current capacity in sectors */
 #define CAPACITY_MSB		58
 #define SECTOR_XFER_CUR		59  /* r/w multiple: current sectors xfered */
@@ -103,6 +103,7 @@
 #define SCT_SUPP		206 /* SMART command transport (SCT) support */
 #define TRANSPORT_MAJOR		222 /* PATA vs. SATA etc.. */
 #define TRANSPORT_MINOR		223 /* minor revision number */
+#define NMRR			217 /* nominal media rotation rate */
 #define INTEGRITY		255 /* integrity word */
 
 /* bit definitions within the words */
@@ -112,8 +113,6 @@
 #define VALID			0xc000
 #define VALID_VAL		0x4000
 /* many words are considered invalid if they are either all-0 or all-1 */
-#define NOVAL_0			0x0000
-#define NOVAL_1			0xffff
 
 /* word 0: gen_config */
 #define NOT_ATA			0x8000	
@@ -121,8 +120,6 @@
 #define MEDIA_REMOVABLE		0x0080
 #define DRIVE_NOT_REMOVABLE	0x0040  /* bit obsoleted in ATA 6 */
 #define INCOMPLETE		0x0004
-#define CFA_SUPPORT_VAL1	0x848a	/* 848a=CFA feature set support */
-#define CFA_SUPPORT_VAL2	0x844a	/* 844a=also means CFA feature set support */
 #define DRQ_RESPONSE_TIME	0x0060
 #define DRQ_3MS_VAL		0x0000
 #define DRQ_INTR_VAL		0x0020
@@ -204,7 +201,7 @@ const char *ata1_cfg_str[] = {			/* word 0 in ATA-1 mode */
 #define MULTIPLE_SETTING_VALID  0x0100  /* 1=multiple sector setting is valid */
 
 /* word 49: capabilities 0 */
-#define STD_STBY  	  	0x2000  /* 1=standard values supported (ATA);
+#define STD_STBY  		0x2000  /* 1=standard values supported (ATA);
 					   0=vendor specific values */
 #define IORDY_SUP		0x0800  /* 1=support; 0=may be supported */
 #define IORDY_OFF		0x0400  /* 1=may be disabled */
@@ -222,7 +219,7 @@ const char *ata1_cfg_str[] = {			/* word 0 in ATA-1 mode */
 #define MODE			0xff00  /* the mode is in the MSBs */
 
 /* word 53: whats_valid */
-#define OK_W88     	   	0x0004	/* the ultra_dma info is valid */
+#define OK_W88     		0x0004	/* the ultra_dma info is valid */
 #define OK_W64_70		0x0002  /* see above for word descriptions */
 #define OK_W54_58		0x0001  /* current cyl, head, sector, cap. info valid */
 
@@ -239,7 +236,7 @@ const char *ata1_cfg_str[] = {			/* word 0 in ATA-1 mode */
 #define DEPTH_BITS		0x001f  /* bits used for queue depth */
 
 /* words 80-81: version numbers */
-/* NOVAL_0 or  NOVAL_1 means device does not report version */
+/* 0x0000 or  0xffff means device does not report version */
 
 /* word 81: minor version number */
 #define MINOR_MAX		0x22
@@ -292,8 +289,8 @@ const char actual_ver[MINOR_MAX+2] = {
 	3,		/* 0x0006	WARNING:   *exactly*		*/
 	2,		/* 0x0007	WARNING:   to the ATA/		*/
 	3,		/* 0x0008	WARNING:   ATAPI version	*/
-	2,		/* 0x0009	WARNING:   listed in	 	*/
-	3,		/* 0x000a	WARNING:   the 		 	*/
+	2,		/* 0x0009	WARNING:   listed in		*/
+	3,		/* 0x000a	WARNING:   the 			*/
 	3,		/* 0x000b	WARNING:   minor_str 		*/
 	3,		/* 0x000c	WARNING:   array		*/
 	4,		/* 0x000d	WARNING:   above.		*/
@@ -332,6 +329,25 @@ static const char unknown[8] = "obsolete";
 //static const char unknown[8] = "unknown";
 #define unknown "unknown-"
 
+static const char *feat_word69_str[16] = { 
+	"CFast specification support",			/* word 69 bit 15 */
+	"Deterministic read data after TRIM",		/* word 69 bit 14 */
+	"Long physical sector diagnostics",		/* word 69 bit 13 */
+	"DEVICE CONFIGURATION SET/IDENTIFY DMA commands", /* word 69 bit 12 */
+	"READ BUFFER DMA command",			/* word 69 bit 11 */
+	"WRITE BUFFER DMA command",			/* word 69 bit 10 */
+	"SET MAX SETPASSWORD/UNLOCK DMA commands",	/* word 69 bit  9 */
+	"DOWNLOAD MICROCODE DMA command",		/* word 69 bit  8 */
+	"reserved 69[7]",				/* word 69 bit  7 */
+	"reserved 69[6]",				/* word 69 bit  6 */
+	"Deterministic read ZEROs after TRIM",		/* word 69 bit  5 */
+	"reserved 69[4]",				/* word 69 bit  4 */
+	"reserved 69[3]",				/* word 69 bit  3 */
+	"reserved 69[2]",				/* word 69 bit  2 */
+	"reserved 69[1]",				/* word 69 bit  1 */
+	"reserved 69[0]",				/* word 69 bit  0 */
+};
+
 static const char *feat_word82_str[16] = { 
 	"obsolete 82[15]",				/* word 82 bit 15: obsolete  */
 	"NOP cmd",					/* word 82 bit 14 */
@@ -408,8 +424,8 @@ static const char *cap_sata0_str[16] = {
 	"unknown 76[15]",				/* word 76 bit 15 */
 	"unknown 76[14]",				/* word 76 bit 14 */
 	"unknown 76[13]",				/* word 76 bit 13 */
-	"unknown 76[12]",				/* word 76 bit 12 */
-	"unknown 76[11]",				/* word 76 bit 11 */
+	"NCQ priority information",			/* word 76 bit 12 */
+	"Idle-Unload when NCQ is active",		/* word 76 bit 11 */
 	"Phy event counters",				/* word 76 bit 10 */
 	"Host-initiated interface power management",	/* word 76 bit  9 */
 	"Native Command Queueing (NCQ)",		/* word 76 bit  8 */
@@ -418,8 +434,8 @@ static const char *cap_sata0_str[16] = {
 	"unknown 76[5]",				/* word 76 bit  5 */
 	"unknown 76[4]",				/* word 76 bit  4 */
 	"unknown 76[3]",				/* word 76 bit  3 */
-	"SATA-II signaling speed (3.0Gb/s)",		/* word 76 bit  2 */
-	"SATA-I signaling speed (1.5Gb/s)",		/* word 76 bit  1 */
+	"Gen2 signaling speed (3.0Gb/s)",		/* word 76 bit  2 */
+	"Gen1 signaling speed (1.5Gb/s)",		/* word 76 bit  1 */
 	"unknown 76[0]"					/* word 76 bit  0 */
 };
 static const char *feat_sata0_str[16] = {
@@ -451,7 +467,7 @@ static const char *feat_sata0_str[16] = {
 #define ERASE_BITS		0x00ff
 
 /* word 92: master password revision */
-/* NOVAL_0 or  NOVAL_1 means no support for master password revision */
+/* 0x0000 or  0xffff means no support for master password revision */
 
 /* word 93: hw reset result */
 #define CBLID			0x2000  /* CBLID status */
@@ -508,7 +524,33 @@ static const char *feat_sct_str[16] = {
 #define SIG_VAL			0x00A5  /* signature value */
 
 __u8 mode_loop(__u16 mode_sup, __u16 mode_sel, int cc, __u8 *have_mode);
-void print_ascii(__u16 *p, __u8 length);
+
+static void print_ascii(__u16 *p, unsigned int length) {
+	__u8 ii;
+	char cl;
+
+	/* find first non-space & print it */
+	for (ii = 0; ii< length; ii++) {
+		if(((char) 0x00ff&((*p)>>8)) != ' ') break;
+		if((cl = (char) 0x00ff&(*p)) != ' ') {
+			if(cl != '\0') printf("%c",cl);
+			p++; ii++;
+			break;
+		}
+		p++;
+	}
+	/* print the rest */
+	for (; ii < length; ii++) {
+		__u8 c;
+		/* some older devices have NULLs */
+		c = (*p) >> 8;
+		if (c) putchar(c);
+		c = (*p);
+		if (c) putchar(c);
+		p++;
+	}
+	printf("\n");
+}
 
 // Given a known-supported ATA major revision,
 // return the lowest possible supported ATA revision.
@@ -563,15 +605,19 @@ static int print_transport_type(__u16 val[])
 			break;
 		case 1:
 			printf("Serial");
-			if (subtype & 0xf) {
-				if (subtype & 1)
+			if (subtype & 0x2f) {
+				if (subtype & (1<<0))
 					printf(", ATA8-AST");
-				if (subtype & 2)
+				if (subtype & (1<<1))
 					printf(", SATA 1.0a");
-				if (subtype & 4)
+				if (subtype & (1<<2))
 					printf(", SATA II Extensions");
-				if (subtype & 8)
+				if (subtype & (1<<3))
 					printf(", SATA Rev 2.5");
+				if (subtype & (1<<4))
+					printf(", SATA Rev 2.6");
+				if (subtype & (1<<5))
+					printf(", SATA Rev 3.0");
 			}
 			break;
 		default:
@@ -592,10 +638,20 @@ static int print_transport_type(__u16 val[])
 	return transport;
 }
 
+static int is_cfa_dev (__u16 *id)
+{
+	/*
+	 * id[0] == 0x848a means "CFA compliant, not ATA-4 compliant".
+	 * id[0] == 0x044a is also allowed, but ISTR that some HDs use it too.
+	 * Also, bit 0x0004 of id[83] means "supports CFA feature set".
+	 */
+	return id[0] == 0x848a || id[0] == 0x844a || (id[83] & 0xc004) == 0x4004;
+}
+
 /* our main() routine: */
 void identify (__u16 *id_supplied)
 {
-
+	unsigned int sector_bytes = 512;
 	__u16 val[256], ii, jj, kk;
 	__u16 like_std = 1, std = 0, min_std = 0xffff;
 	__u16 dev = NO_DEV, eqpt = NO_DEV;
@@ -615,14 +671,15 @@ void identify (__u16 *id_supplied)
 	/* check if we recognise the device type */
 	printf("\n");
 
-	if(!(val[GEN_CONFIG] & NOT_ATA)) {
-		dev = ATA_DEV;
-		printf("ATA device, with ");
-	} else if(val[GEN_CONFIG]==CFA_SUPPORT_VAL1 || val[GEN_CONFIG]==CFA_SUPPORT_VAL2) {
+	//if(val[GEN_CONFIG] == 0x848a || val[GEN_CONFIG] == 0x844a) {
+	if (is_cfa_dev(val)) {
 		is_cfa = 1;
 		dev = ATA_DEV;
 		like_std = 4;
 		printf("CompactFlash ATA device\n");
+	} else if(!(val[GEN_CONFIG] & NOT_ATA)) {
+		dev = ATA_DEV;
+		printf("ATA device, with ");
 	} else if(!(val[GEN_CONFIG] & NOT_ATAPI)) {
 		dev = ATAPI_DEV;
 		eqpt = (val[GEN_CONFIG] & EQPT_TYPE) >> SHIFT_EQPT;
@@ -647,7 +704,7 @@ void identify (__u16 *id_supplied)
 	 */
 	if((val[CONFIG]==STBY_NID_VAL) || (val[CONFIG]==STBY_ID_VAL) ||
 	   (val[CONFIG]==PWRD_NID_VAL) || (val[CONFIG]==PWRD_ID_VAL) ) {
-	   	like_std = 5;
+		like_std = 5;
 		if((val[CONFIG]==STBY_NID_VAL) || (val[CONFIG]==STBY_ID_VAL))
 			printf("powers-up in standby; SET FEATURES subcmd spins-up.\n");
 		if(((val[CONFIG]==STBY_NID_VAL) || (val[CONFIG]==PWRD_NID_VAL)) &&
@@ -712,7 +769,7 @@ void identify (__u16 *id_supplied)
 		/* looks like when they up-issue the std, they obsolete one;
 		 * thus, only the newest 4 issues need be supported.
 		 * (That's what "kk" and "min_std" are all about) */
-		if(val[MAJOR] && (val[MAJOR] != NOVAL_1)) {
+		if(val[MAJOR] && (val[MAJOR] != 0xffff)) {
 			printf("\n\tSupported: ");
 			jj = val[MAJOR] << 1;
 			kk = min_ata_std(like_std);
@@ -724,7 +781,7 @@ void identify (__u16 *id_supplied)
 						kk = min_ata_std(like_std);
 					}
 					if (min_std > ii)
-					       	min_std = ii;
+						min_std = ii;
 				}
 				jj <<= 1;
 			}
@@ -774,12 +831,12 @@ void identify (__u16 *id_supplied)
 		} else  printf("\n");
 	} else {
 		/* TBD: do CDROM stuff more thoroughly.  For now... */
-	  	kk = 0;
+		kk = 0;
 		if(val[CDR_MINOR] == 9) {
 			kk = 1;
 			printf("\n\tUsed: ATAPI for CD-ROMs, SFF-8020i, r2.5");
 		}
-		if(val[CDR_MAJOR] && (val[CDR_MAJOR] != NOVAL_1)) {
+		if(val[CDR_MAJOR] && (val[CDR_MAJOR] != 0xffff)) {
 			kk = 1;
 			printf("\n\tSupported: CD-ROM ATAPI");
 			jj = val[CDR_MAJOR] >> 1;
@@ -823,12 +880,12 @@ void identify (__u16 *id_supplied)
 		default : printf("Unknown\n"); break;
 		}
 	} else {
-	  	/* addressing...CHS? See section 6.2 of ATA specs 4 or 5 */
+		/* addressing...CHS? See section 6.2 of ATA specs 4 or 5 */
 		ll = 0; mm = 0; bb = 0; bbbig = 0;
 		if (val[CAPAB_0] & LBA_SUP)
 			ll = (__u32)val[LBA_SECTS_MSB] << 16 | val[LBA_SECTS_LSB];
-	  	if ( (ll > 0x00FBFC10) && (!val[LCYLS])) {
-		  	printf("\tCHS addressing not supported\n");
+		if ( (ll > 0x00FBFC10) && (!val[LCYLS])) {
+			printf("\tCHS addressing not supported\n");
 		} else {
 			jj = val[WHATS_VALID] & OK_W54_58;
 			printf("\tLogical\t\tmax\tcurrent\n");
@@ -868,8 +925,25 @@ void identify (__u16 *id_supplied)
 				printf("\tLBA48  user addressable sectors:%11llu\n", (unsigned long long)bbbig);
 			}
 		}
+		if((val[106] & 0xc000) != 0x4000) {
+			printf("\t%-31s %11u bytes\n","Logical/Physical Sector size:", sector_bytes);
+		} else {
+			unsigned int lsize = 256, pfactor = 1;
+			if (val[106] & (1<<13))
+				pfactor = (1 << (val[106] & 0xf));
+			if (val[106] & (1<<12))
+				lsize = (val[118] << 16) | val[117];
+			sector_bytes = 2 * lsize;
+			printf("\t%-31s %11u bytes\n","Logical  Sector size:", sector_bytes);
+			printf("\t%-31s %11u bytes\n","Physical Sector size:", sector_bytes * pfactor);
+			if ((val[209] & 0xc000) == 0x4000) {
+				unsigned int offset = val[209] & 0x1fff;
+				printf("\t%-31s %11u bytes\n", "Logical Sector-0 offset:", offset * lsize);
+			}
+		}
 		if (!bbbig) bbbig = (__u64)(ll>mm ? ll : mm); /* # 512 byte blocks */
 		if (!bbbig) bbbig = bb;
+		bbbig *= (sector_bytes / 512);
 		printf("\tdevice size with M = 1024*1024: %11llu MBytes\n", (unsigned long long)(bbbig>>11));
 		bbbig = (bbbig<<9)/1000000;
 		printf("\tdevice size with M = 1000*1000: %11llu MBytes ", (unsigned long long)bbbig);
@@ -879,8 +953,8 @@ void identify (__u16 *id_supplied)
 
 	/* device cache/buffer size, if reported (obsolete field, but usually valid regardless) */
 	printf("\tcache/buffer size  = ");
-	if (val[20] <= 3 && val[21] && val[21] != 0xffff) {
-		printf("%u KBytes", val[21] / 2);
+	if (val[20] <= 3 && val[BUF_SIZE] && val[BUF_SIZE] != 0xffff) {
+		printf("%u KBytes", val[BUF_SIZE] / 2);
 		if (val[20])
 			printf(" (type=%s)", BuffType[val[20]]);
 	} else {
@@ -888,6 +962,38 @@ void identify (__u16 *id_supplied)
 	}
 	putchar('\n');
 
+	/* Form factor */
+	if(val[168] && (val[168] & 0xfff8) == 0) {
+		printf("\tForm Factor: ");
+		switch(val[168]) {
+		case 1:
+			printf("5.25 inch");
+			break;
+		case 2:
+			printf("3.5 inch");
+			break;
+		case 3:
+			printf("2.5 inch");
+			break;
+		case 4:
+			printf("1.8 inch");
+			break;
+		case 5:
+			printf("less than 1.8 inch");
+			break;
+		default:
+			printf("unknown (0x%04x]", val[168]);
+			break;
+		}
+		printf("\n");
+	}
+
+	/* Spinning disk or solid state? */
+	if(val[NMRR] == 1)
+		printf("\tNominal Media Rotation Rate: Solid State Device\n");
+	else if(val[NMRR] > 0x401)
+		printf("\tNominal Media Rotation Rate: %u\n", val[NMRR]);
+
 	/* hw support of commands (capabilities) */
 	printf("Capabilities:\n");
 	printf("\t");
@@ -917,7 +1023,7 @@ void identify (__u16 *id_supplied)
 		printf("\n");
 	}
 	jj = 0;
-	if((min_std == 1) && (val[BUF_SIZE] && (val[BUF_SIZE] != NOVAL_1))) {
+	if((min_std == 1) && (val[BUF_SIZE] && (val[BUF_SIZE] != 0xffff))) {
 		printf("\tBuffer size: %.1fkB",(float)val[BUF_SIZE]/2);
 		jj = 1;
 	}
@@ -950,7 +1056,7 @@ void identify (__u16 *id_supplied)
 			if(val[CAPAB_0] & STD_STBY) printf("Standard");
 			else 			    printf("Vendor");
 			if((like_std > 3) && ((val[CAPAB_1] & VALID) == VALID_VAL)) {
-			   	if(val[CAPAB_1] & MIN_STANDBY_TIMER) printf(", with ");
+				if(val[CAPAB_1] & MIN_STANDBY_TIMER) printf(", with ");
 				else 				     printf(", no ");
 				printf("device specific minimum\n");
 			} else  printf("\n");
@@ -965,7 +1071,7 @@ void identify (__u16 *id_supplied)
 				printf("%u\n",val[SECTOR_XFER_CUR] & SECTOR_XFER);
 			else	printf("?\n");
 		}
-		if((like_std > 3) && (val[CMDS_SUPP_1] & 0x0008)) {
+		if((like_std > 3) && (val[CMDS_SUPP_1] & 0xc008) == 0x4008) {
 			printf("\tAdvanced power management level: ");
 			if (val[CMDS_EN_1] & 0x0008)
 				printf("%u\n", val[ADV_PWR] & 0xff);
@@ -1098,63 +1204,21 @@ void identify (__u16 *id_supplied)
 		if (val[SCT_SUPP] & 0x1)
 			print_features(val[SCT_SUPP], val[SCT_SUPP] & 0x3f, feat_sct_str);
 	}
-	if((val[RM_STAT] & RM_STAT_BITS) == RM_STAT_SUP) 
-		printf("\tRemovable Media Status Notification feature set supported\n");
-
-	/* security */
-	if((eqpt != CDROM) && (like_std > 3) && (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME]))
-	{
-		printf("Security: \n");
-		if(val[PSWD_CODE] && (val[PSWD_CODE] != NOVAL_1))
-			printf("\tMaster password revision code = %u\n",val[PSWD_CODE]);
-		jj = val[SECU_STATUS];
-		if(jj) {
-			for (ii = 0; ii < NUM_SECU_STR; ii++) {
-				if(!(jj & 0x0001)) printf("\tnot\t");
-				else		   printf("\t\t");
-				printf("%s\n",secu_str[ii]);
-				jj >>=1;
-			}
-			if(val[SECU_STATUS] & SECU_ENABLED) {
-				printf("\tSecurity level ");
-				if(val[SECU_STATUS] & SECU_LEVEL) printf("maximum\n");
-				else				  printf("high\n");
+	if (like_std > 6) {
+		const __u16 trimd = 1<<14;	/* deterministic read data after TRIM */
+		const __u16 trimz = 1<<5;	/* deterministic read ZEROs after TRIM */
+		__u16 word69 = val[69] & ~(trimz | trimd); /* TRIM bits require special interpretation */
+		print_features(word69, word69, feat_word69_str);
+		if (val[169] & 1 && val[169] != 0xffff) { /* supports TRIM ? */
+			printf("\t   *\tData Set Management TRIM supported\n");
+			if (val[69] & trimd) { /* Deterministic TRIM support */
+				if (val[69] & trimz)
+					print_features(trimz, trimz, feat_word69_str);
+				else
+					print_features(trimd, trimd, feat_word69_str);
 			}
 		}
-		jj =  val[ERASE_TIME]     & ERASE_BITS;
-		kk =  val[ENH_ERASE_TIME] & ERASE_BITS;
-		if(jj || kk) {
-			printf("\t");
-			if(jj) printf("%umin for SECURITY ERASE UNIT. ", jj==ERASE_BITS ? 508 : jj<<1);
-			if(kk) printf("%umin for ENHANCED SECURITY ERASE UNIT.", kk==ERASE_BITS ? 508 : kk<<1);
-			printf("\n");
-		}
-	}
-	if((eqpt != CDROM) && (like_std > 3) && (val[CMDS_EN_2] & WWN_SUP)) 
-    {
-		printf("Logical Unit WWN Device Identifier: %04x%04x%04x%04x\n", val[108], val[109], val[110], val[111]);
-		printf("\tNAA\t\t: %x\n", (val[108] & 0xf000) >> 12);
-		printf("\tIEEE OUI\t: %06x\n", (((val[108] & 0x0fff) << 12) | ((val[109] & 0xfff0) >> 4)));
-		printf("\tUnique ID\t: %x%08x\n", (val[109] & 0x000f), ((val[110] << 16) | val[111]));
-    }
-
-	/* reset result */
-	if((val[HWRST_RSLT] & VALID) == VALID_VAL) {
-		printf("HW reset results:\n");
-		if(val[HWRST_RSLT] & CBLID) printf("\tCBLID- above Vih\n");
-		else			    printf("\tCBLID- below Vih\n");
-		if(val[HWRST_RSLT] & RST0)  {
-			printf("\tDevice num = 0");
-			jj = val[HWRST_RSLT];
-		} else {
-			printf("\tDevice num = 1");
-			jj = val[HWRST_RSLT] >> 8;
-		}
-		if((jj & DEV_DET) == JUMPER_VAL) 
-			printf(" determined by the jumper");
-		else if((jj & DEV_DET) == CSEL_VAL)
-			printf(" determined by CSEL");
-		printf("\n");
+		
 	}
 
 	if (is_cfa) {
@@ -1179,7 +1243,7 @@ void identify (__u16 *id_supplied)
 			for (mode = 1; mode <= max; ++mode) {
 				if (mode == selected)
 					strcat(modes, "*");
-				sprintf(modes + strlen(modes), "mdma%u ", mode + 4);
+				sprintf(modes + strlen(modes), "mdma%u ", mode + 2);
 			}
 		}
 		if (val[164] & 0x8000)
@@ -1217,6 +1281,7 @@ void identify (__u16 *id_supplied)
 			printf("\t   *\tCFA advanced modes: %s\n", modes);
 
 		if(val[CFA_PWR_MODE] & VALID_W160) {
+			putchar('\t');
 			if((val[CFA_PWR_MODE] & PWR_MODE_REQ) == 0)
 				printf("   *");
 			printf("\tCFA Power Level 1 ");
@@ -1233,6 +1298,64 @@ void identify (__u16 *id_supplied)
 		}
 	}
 
+	if((val[RM_STAT] & RM_STAT_BITS) == RM_STAT_SUP) 
+		printf("\t\tRemovable Media Status Notification feature set supported\n");
+
+	/* security */
+	if((eqpt != CDROM) && (like_std > 3) && (val[SECU_STATUS] || val[ERASE_TIME] || val[ENH_ERASE_TIME]))
+	{
+		printf("Security: \n");
+		if(val[PSWD_CODE] && (val[PSWD_CODE] != 0xffff))
+			printf("\tMaster password revision code = %u\n",val[PSWD_CODE]);
+		jj = val[SECU_STATUS];
+		if(jj) {
+			for (ii = 0; ii < NUM_SECU_STR; ii++) {
+				if(!(jj & 0x0001)) printf("\tnot\t");
+				else		   printf("\t\t");
+				printf("%s\n",secu_str[ii]);
+				jj >>=1;
+			}
+			if(val[SECU_STATUS] & SECU_ENABLED) {
+				printf("\tSecurity level ");
+				if(val[SECU_STATUS] & SECU_LEVEL) printf("maximum\n");
+				else				  printf("high\n");
+			}
+		}
+		jj =  val[ERASE_TIME]     & ERASE_BITS;
+		kk =  val[ENH_ERASE_TIME] & ERASE_BITS;
+		if(jj || kk) {
+			printf("\t");
+			if(jj) printf("%umin for SECURITY ERASE UNIT. ", jj==ERASE_BITS ? 508 : jj<<1);
+			if(kk) printf("%umin for ENHANCED SECURITY ERASE UNIT.", kk==ERASE_BITS ? 508 : kk<<1);
+			printf("\n");
+		}
+	}
+	if((eqpt != CDROM) && (like_std > 3) && (val[CMDS_EN_2] & WWN_SUP)) {
+		printf("Logical Unit WWN Device Identifier: %04x%04x%04x%04x\n", val[108], val[109], val[110], val[111]);
+		printf("\tNAA\t\t: %x\n", (val[108] & 0xf000) >> 12);
+		printf("\tIEEE OUI\t: %06x\n", (((val[108] & 0x0fff) << 12) | ((val[109] & 0xfff0) >> 4)));
+		printf("\tUnique ID\t: %x%08x\n", (val[109] & 0x000f), ((val[110] << 16) | val[111]));
+	}
+
+	/* reset result */
+	if((val[HWRST_RSLT] & VALID) == VALID_VAL) {
+		printf("HW reset results:\n");
+		if(val[HWRST_RSLT] & CBLID) printf("\tCBLID- above Vih\n");
+		else			    printf("\tCBLID- below Vih\n");
+		if(val[HWRST_RSLT] & RST0)  {
+			printf("\tDevice num = 0");
+			jj = val[HWRST_RSLT];
+		} else {
+			printf("\tDevice num = 1");
+			jj = val[HWRST_RSLT] >> 8;
+		}
+		if((jj & DEV_DET) == JUMPER_VAL) 
+			printf(" determined by the jumper");
+		else if((jj & DEV_DET) == CSEL_VAL)
+			printf(" determined by CSEL");
+		printf("\n");
+	}
+
 	/* more stuff from std 5 */
 	if ((like_std > 4) && (eqpt != CDROM)) {
 		if ((val[INTEGRITY] & SIG) == SIG_VAL) {
@@ -1263,33 +1386,6 @@ __u8 mode_loop(__u16 mode_sup, __u16 mode_sel, int cc, __u8 *have_mode) {
 	return err_dma;
 }
 
-void print_ascii(__u16 *p, __u8 length) {
-	__u8 ii;
-	char cl;
-	
-	/* find first non-space & print it */
-	for (ii = 0; ii< length; ii++) {
-		if(((char) 0x00ff&((*p)>>8)) != ' ') break;
-		if((cl = (char) 0x00ff&(*p)) != ' ') {
-			if(cl != '\0') printf("%c",cl);
-			p++; ii++;
-			break;
-		}
-		p++;
-	}
-	/* print the rest */
-	for (; ii < length; ii++) {
-		__u8 c;
-		/* some older devices have NULLs */
-		c = (*p) >> 8;
-		if (c) putchar(c);
-		c = (*p);
-		if (c) putchar(c);
-		p++;
-	}
-	printf("\n");
-}
-
 void dco_identify_print (__u16 *dco)
 {
 	__u64 lba;
