@@ -3,21 +3,25 @@
 #include <linux/types.h>
 
 enum {
+	ATA_OP_DSM			= 0x06, // Data Set Management (TRIM)
 	ATA_OP_READ_PIO			= 0x20,
 	ATA_OP_READ_PIO_ONCE		= 0x21,
 	ATA_OP_READ_LONG		= 0x22,
 	ATA_OP_READ_LONG_ONCE		= 0x23,
 	ATA_OP_READ_PIO_EXT		= 0x24,
 	ATA_OP_READ_DMA_EXT		= 0x25,
+	ATA_OP_READ_FPDMA		= 0x60,	// NCQ
 	ATA_OP_WRITE_PIO		= 0x30,
 	ATA_OP_WRITE_LONG		= 0x32,
 	ATA_OP_WRITE_LONG_ONCE		= 0x33,
 	ATA_OP_WRITE_PIO_EXT		= 0x34,
 	ATA_OP_WRITE_DMA_EXT		= 0x35,
+	ATA_OP_WRITE_FPDMA		= 0x61,	// NCQ
 	ATA_OP_READ_VERIFY		= 0x40,
 	ATA_OP_READ_VERIFY_ONCE		= 0x41,
 	ATA_OP_READ_VERIFY_EXT		= 0x42,
 	ATA_OP_WRITE_UNC_EXT		= 0x45,	// lba48, no data, uses feat reg
+	ATA_OP_FORMAT_TRACK		= 0x50,
 	ATA_OP_DOWNLOAD_MICROCODE	= 0x92,
 	ATA_OP_STANDBYNOW2		= 0x94,
 	ATA_OP_CHECKPOWERMODE2		= 0x98,
@@ -27,6 +31,7 @@ enum {
 	ATA_OP_READ_NATIVE_MAX_EXT	= 0x27,
 	ATA_OP_SMART			= 0xb0,
 	ATA_OP_DCO			= 0xb1,
+	ATA_OP_ERASE_SECTORS		= 0xc0,
 	ATA_OP_READ_DMA			= 0xc8,
 	ATA_OP_WRITE_DMA		= 0xca,
 	ATA_OP_DOORLOCK			= 0xde,
@@ -36,6 +41,7 @@ enum {
 	ATA_OP_SETIDLE			= 0xe3,
 	ATA_OP_SET_MAX			= 0xf9,
 	ATA_OP_SET_MAX_EXT		= 0x37,
+	ATA_OP_SET_MULTIPLE		= 0xc6,
 	ATA_OP_CHECKPOWERMODE1		= 0xe5,
 	ATA_OP_SLEEPNOW1		= 0xe6,
 	ATA_OP_FLUSHCACHE		= 0xe7,
@@ -109,26 +115,33 @@ enum {
 };
 
 union reg_flags {
-	unsigned all			:16;
-	struct {
-		unsigned data		: 1;
-		unsigned feat		: 1;
-		unsigned lbal		: 1;
-		unsigned nsect		: 1;
-		unsigned lbam		: 1;
-		unsigned lbah		: 1;
-		unsigned dev		: 1;
-		unsigned command	: 1;
-
-		unsigned hob_data	: 1;
-		unsigned hob_feat	: 1;
-		unsigned hob_lbal	: 1;
-		unsigned hob_nsect	: 1;
-		unsigned hob_lbam	: 1;
-		unsigned hob_lbah	: 1;
-		unsigned hob_dev	: 1;
-		unsigned hob_command	: 1;
-	} b;
+	unsigned all				: 16;
+	union {
+		unsigned lob_all		: 8;
+		struct {
+			unsigned data		: 1;
+			unsigned feat		: 1;
+			unsigned lbal		: 1;
+			unsigned nsect		: 1;
+			unsigned lbam		: 1;
+			unsigned lbah		: 1;
+			unsigned dev		: 1;
+			unsigned command	: 1;
+		} lob;
+	};
+	union {
+		unsigned hob_all		: 8;
+		struct {
+			unsigned data		: 1;
+			unsigned feat		: 1;
+			unsigned lbal		: 1;
+			unsigned nsect		: 1;
+			unsigned lbam		: 1;
+			unsigned lbah		: 1;
+			unsigned dev		: 1;
+			unsigned command	: 1;
+		} hob;
+	};
 };
 
 struct taskfile_regs {
@@ -198,6 +211,9 @@ struct scsi_sg_io_hdr {
 #define SG_ATA_16		0x85
 #define SG_ATA_16_LEN		16
 
+#define SG_ATA_12		0xa1
+#define SG_ATA_12_LEN		12
+
 #define SG_ATA_LBA48		1
 #define SG_ATA_PROTO_NON_DATA	( 3 << 1)
 #define SG_ATA_PROTO_PIO_IN	( 4 << 1)
